/*
 * ******************************************************************************************
 *
 * File		: main.c
 * Author	: GOWIN
 * Function	: Main.
 *
 * ******************************************************************************************
 */


// Includes ---------------------------------------------------------------------------------
#include "apb_i2c_dev.h"
#include "uart.h"
#include <stdio.h>


// Definitions ------------------------------------------------------------------------------

// Select I2C master mode or slave mode
// For user defined
#define I2C_MASTER_MODE_TEST         1							// Enable/disable I2C master mode
#define I2C_SLAVE_MODE_TEST          !(I2C_MASTER_MODE_TEST)	// Enable/disable I2C slave mode

// Select I2C 10bit or 7bit address mode
// For user defined
#define I2C_10BIT_MODE               0							// Enable/disable I2C 10bit address mode
#define I2C_7BIT_MODE                !(I2C_10BIT_MODE)			// Enable/disable I2C 7bit address mode


// Simulate I2C slave EEPROM device
// Simulate 10 bit I2C slave address
// User defined
#if I2C_10BIT_MODE
#define SIM_EEPROM_I2C_ADDR          (0x3FA)
#else
#define SIM_EEPROM_I2C_ADDR          (0x60)
#endif

#define I2C_SLAVE_ADDR               SIM_EEPROM_I2C_ADDR

// Simulate I2C EEPROM offset address 0x0 ~ 0x400
// Max memory locations available
#define EEPROM_MAX_ADDR              1024

// Max bytes to write in one step
#define EEPROM_MAX_WRITE             16


#define MEMSET(s, c, n)              __builtin_memset ((s), (c), (n))
#define MEMCPY(des, src, n)          __builtin_memcpy((des), (src), (n))

#define FAKE_CPUFREQ_S		PCLKFREQ		// PCLK
#define SLAVE_LATENCY_MS	100
#define DELAY_CYCLE		    (SLAVE_LATENCY_MS * (FAKE_CPUFREQ_S/1000))

// I2C slave TX/RX flag
typedef enum _SLAVE_TRX_FLAG_ID
{
	I2C_SLAVE_TX_NONE = 0,
	I2C_SLAVE_RX_NONE = 0,
	I2C_SLAVE_TX_START,
	I2C_SLAVE_TX_COMPLETE,
	I2C_SLAVE_RX_START,
	I2C_SLAVE_RX_COMPLETE,
} SLAVE_TRX_FLAG_ID;


/*************************************** Master ****************************************/
#if I2C_MASTER_MODE_TEST
// Test I2C master mode

static unsigned int Device_Addr = I2C_SLAVE_ADDR;
static unsigned char wr_buf[EEPROM_MAX_WRITE + 2] = {0};

// 2 byte address since EEPROM_MAX_ADDR is 1024
static volatile unsigned int EEPROM_Addr = 0x0216;

// 10 byte data
static volatile unsigned char EEPROM_Data[10] = {0x12, 0x34, 0x56, 0x78, 0x9A, 0xBC, 0xDE, 0xF0, 0x5A, 0xA5};
static volatile unsigned char Tmp[10] = {0};

// Master transmit
int i2c_master_tx(unsigned int addr, const unsigned char* buf, unsigned int len)
{
	// 2 bytes for simulate I2C EEPROM offset address 0x0 ~ 0x400
	wr_buf[0] = (unsigned char)(addr >> 8);
	wr_buf[1] = (unsigned char)(addr & 0xFF);

	MEMCPY(&wr_buf[2], &buf[0], len);

	apb_i2c1_master_transmit(Device_Addr, wr_buf, len + 2, false);
	while(apb_i2c1_get_status().busy);

	if(apb_i2c1_get_datacount() != (len + 2))
	{
		return -1;
	}

	return 0;
}

// Master receive
int i2c_master_rx(unsigned int addr, unsigned char* buf, unsigned int len)
{
	unsigned char a[2] = {0};

	a[0] = (unsigned char)(addr >> 8);
	a[1] = (unsigned char)(addr & 0xFF);

	// xfer_pending is true => no stop condition, still busy
	// since xfer_pending is true, no stop condition and busy bit will always be pull high
	// 10-bit slave address must set STOP bit
	// tell EEPROM which EEPROM offset address will be read
	apb_i2c1_master_transmit(Device_Addr, a, 2, true);
	while(apb_i2c1_get_status().busy);

	if(apb_i2c1_get_datacount() != 2)
	{
		return -1;
	}

	// Read-request to read the data from the EEPROM offset address
	apb_i2c1_master_receive(Device_Addr, buf, len, false);
	while(apb_i2c1_get_status().busy);

	if(apb_i2c1_get_datacount() != len)
	{
		return -1;
	}

	return 0;
}

// Initializes I2C master mode
int i2c_master_init(void)
{
	int status = APB_I2C_DRIVER_OK;

	// Default slave mode
	status = apb_i2c1_initialize(NULL);
	if(status != APB_I2C_DRIVER_OK)
	{
		return status;
	}

	// Clear
	status = apb_i2c1_control(APB_I2C_BUS_CLEAR, 0);
	if(status != APB_I2C_DRIVER_OK)
	{
		return status;
	}

	// Power full
	status = apb_i2c1_power_control(APB_I2C_POWER_FULL);
	if(status != APB_I2C_DRIVER_OK)
	{
		return status;
	}

	// Speed: standard
	status = apb_i2c1_control(APB_I2C_BUS_SPEED, APB_I2C_BUS_SPEED_STANDARD);
	if(status != APB_I2C_DRIVER_OK)
	{
		return status;
	}

	Device_Addr = I2C_SLAVE_ADDR;

	// Simulate I2C slave EEPROM device
	// Simulate I2C slave address
#if I2C_10BIT_MODE
	status = apb_i2c1_control(APB_I2C_OWN_ADDRESS, (Device_Addr | APB_I2C_ADDRESS_10BIT));
#else
	status = apb_i2c1_control(APB_I2C_OWN_ADDRESS, (Device_Addr));
#endif
	if(status != APB_I2C_DRIVER_OK)
	{
		return status;
	}

	return status;
}

// Master mode
int i2c_master_mode(void)
{
	volatile unsigned int time_start = 0;
	volatile unsigned int time_end = 0;
	int status = 0, i = 0;

	// Initializes UART
	uart_init(38400);	// Baud rate is 38400

	printf("\r\nIt's an APB I2C Master Mode demo.\r\n");
	printf("\r\n====Please hardwire 2 test board=====\r\n\r\n");


	printf("I2C master mode init .....\r\n");

	status = i2c_master_init();
	if(status != APB_I2C_DRIVER_OK)
	{
		printf("I2C init error.....\r\n");
		while(1);
	}

	printf("I2C FIFO master tx .....\r\n");

	status = i2c_master_tx(EEPROM_Addr, (unsigned char*)&EEPROM_Data, sizeof(EEPROM_Data));

	time_start = DEV_PLMT->MTIME[0];			//
	time_end = DEV_PLMT->MTIME[0];				//

	printf("I2C FIFO master rx .....\r\n");

	// Waiting slave to complete the sending setting
	while (time_end < (time_start + DELAY_CYCLE))
	{
		time_end = DEV_PLMT->MTIME[0];			//
	}

	status = i2c_master_rx(EEPROM_Addr, (unsigned char*)&Tmp, sizeof(Tmp));

	for(i = 0; i < 10; i++)
	{
		// Error hit
		if(EEPROM_Data[i] != Tmp[i])
		{
			printf("I2C FIFO master test fail.....\r\n");
			while(1);
		}
	}

	printf("I2C FIFO master test pass .....\r\n");

	// Exit demo
	apb_i2c1_uninitialize();


	return 0;
}

/*************************************** Master ****************************************/


/*************************************** Slave ****************************************/
#elif I2C_SLAVE_MODE_TEST
// Test I2C slave mode

// Simulate I2C slave EEPROM device
static volatile unsigned int offset_addr = 0;
static volatile unsigned char slave_eeprom[EEPROM_MAX_ADDR] = {0};
static volatile int slave_tx = I2C_SLAVE_TX_NONE;
static volatile int slave_rx = I2C_SLAVE_RX_NONE;

// Callback function
void i2c_callback(unsigned int event)
{
	// Callback function body
	if(event & APB_I2C_EVENT_TRANSFER_INCOMPLETE)
	{
		if(event & APB_I2C_EVENT_SLAVE_RECEIVE)
		{
			// Slave address hit is the entry for slave mode driver
			// to detect slave RX/TX action depend on master TX/RX action
			slave_rx = I2C_SLAVE_RX_START;
		}
		else if(event & APB_I2C_EVENT_SLAVE_TRANSMIT)
		{
			// Slave address hit is the entry for slave mode driver
			// to detect slave RX/TX action depend on master TX/RX action
			slave_tx = I2C_SLAVE_TX_START;
		}
	}
	else if(event & APB_I2C_EVENT_TRANSFER_DONE)
	{
		if(slave_rx == I2C_SLAVE_RX_START)
		{
			slave_rx = I2C_SLAVE_RX_COMPLETE;
		}

		if(slave_tx == I2C_SLAVE_TX_START)
		{
			slave_tx = I2C_SLAVE_TX_COMPLETE;
		}
	}
	else if(event & APB_I2C_EVENT_ADDRESS_NACK)
	{
		if(slave_tx == I2C_SLAVE_TX_START)
		{
			slave_tx = I2C_SLAVE_TX_COMPLETE;
		}
	}
}

// Slave transmit
int i2c_slave_tx(void)
{
	unsigned int r_addr = offset_addr;

	do
	{
		apb_i2c1_slave_transmit((unsigned char*)&slave_eeprom[r_addr], 1);
		r_addr++;

		while(apb_i2c1_get_status().busy);

		// Error hit
		if((r_addr - offset_addr) > MAX_XFER_SZ)
		{
			while(1);
		}
	}while(slave_tx != I2C_SLAVE_TX_COMPLETE);

	return 0;
}

// Slave receive
int i2c_slave_rx(void)
{
	unsigned int data_count = 0;
	unsigned char tmp[2] = {0};
	unsigned int w_addr = 0;

	// Master device issue write-request w/ flash address,
	// then issue read-request to read flash data from the address
	offset_addr = 0;

	// A new I2C data transaction(start-address-data-stop)
	data_count = apb_i2c1_get_datacount();

	// Error hit
	if(data_count > MAX_XFER_SZ)
	{
		while(1);
	}

	if(data_count > 0)
	{
		if(data_count >= 2)
		{
			// The first 2 bytes are flash address
			// High byte
			apb_i2c1_slave_receive(&tmp[0], 1);
			// Low byte
			apb_i2c1_slave_receive(&tmp[1], 1);

			w_addr = offset_addr = ((tmp[0] << 8) | tmp[1]);
			apb_i2c1_slave_receive((unsigned char*)&slave_eeprom[w_addr], data_count - 2);
			w_addr += (data_count - 2);
		}
		else
		{
			apb_i2c1_slave_receive((unsigned char*)&slave_eeprom[0], data_count);
		}
	}
	// I2C data transaction w/o pay load data
	else
	{
	}

	return 0;
}

// Initializes I2C slave mode
int i2c_slave_init(void)
{
	int status = APB_I2C_DRIVER_OK;

	// Default slave mode
	status = apb_i2c1_initialize(i2c_callback);
	if(status != APB_I2C_DRIVER_OK)
	{
		return status;
	}

	// Clear
	status = apb_i2c1_control(APB_I2C_BUS_CLEAR, 0);
	if(status != APB_I2C_DRIVER_OK)
	{
		return status;
	}

	// Power full
	status = apb_i2c1_power_control(APB_I2C_POWER_FULL);
	if(status != APB_I2C_DRIVER_OK)
	{
		return status;
	}

	// Speed: standard
	status = apb_i2c1_control(APB_I2C_BUS_SPEED, APB_I2C_BUS_SPEED_STANDARD);
	if(status != APB_I2C_DRIVER_OK)
	{
		return status;
	}

	// Simulate I2C slave EEPROM device
	// Simulate I2C slave address
#if I2C_10BIT_MODE
	status = apb_i2c1_control(APB_I2C_OWN_ADDRESS, (I2C_SLAVE_ADDR | APB_I2C_ADDRESS_10BIT));
#else
	status = apb_i2c1_control(APB_I2C_OWN_ADDRESS, (I2C_SLAVE_ADDR));
#endif
	if(status != APB_I2C_DRIVER_OK)
	{
		return status;
	}

	return status;
}

// Slave mode
int i2c_slave_mode(void)
{
	int status = 0;

	// Initializes UART
	uart_init(38400);	// Baud rate is 38400

	printf("\r\nIt's an APB I2C Slave Mode demo.\r\n");
	printf("\r\n====Please hardwire 2 test board=====\r\n\r\n");


	printf("I2C slave mode init .....\r\n");

	status = i2c_slave_init();
	if(status != APB_I2C_DRIVER_OK)
	{
		while(1);
	}

	do
	{
		if(slave_rx == I2C_SLAVE_RX_COMPLETE)
		{
			// A new I2C data transaction(start-address-data-stop)
			i2c_slave_rx();
			slave_rx = I2C_SLAVE_RX_NONE;
		}

		if(slave_tx == I2C_SLAVE_TX_START)
		{
			// A new I2C data transaction(start-address-data-stop)
			i2c_slave_tx();
			slave_tx = I2C_SLAVE_TX_NONE;
		}
	}while(1);

	return 0;
}
#endif

/*************************************** Slave ****************************************/


// Application entry function
int main(void)
{
#if I2C_MASTER_MODE_TEST
	// Master mode
	i2c_master_mode();
#elif I2C_SLAVE_MODE_TEST
	// Slave mode
	i2c_slave_mode();
#endif

	return 0;
}
